# AUTOGENERATED! DO NOT EDIT! File to edit: ../transformer.ipynb.

# %% auto 0
__all__ = ['lr', 'weight_decay', 'p', 'd_model', 'fn_name', 'frac_train', 'num_epochs', 'save_models', 'save_every',
           'stopping_thresh', 'seed', 'num_layers', 'batch_style', 'd_vocab', 'n_ctx', 'd_mlp', 'num_heads', 'd_head',
           'act_type', 'use_ln', 'random_answers', 'fns_dict', 'fn', 'train', 'test', 'HookPoint', 'Embed', 'Unembed',
           'PosEmbed', 'LayerNorm', 'Attention', 'MLP', 'TransformerBlock', 'Transformer', 'gen_train_test',
           'train_model']

# %% ../transformer.ipynb 4
import torch.nn as nn

class HookPoint(nn.Module):
    '''A helper class to get access to intermediate activations (inspired by Garcon)
    It's a dummy module that is the identity function by default
    I can wrap any intermediate activation in a HookPoint and get a convenient way to add PyTorch hooks
    '''
    def __init__(self):
        super().__init__()
        self.fwd_hooks = []
        self.bwd_hooks = []
    
    def give_name(self, name):
        # Called by the model at initialisation
        self.name = name
    
    def add_hook(self, hook, dir='fwd'):
        # Hook format is fn(activation, hook_name)
        # Change it into PyTorch hook format (this includes input and output, 
        # which are the same for a HookPoint)
        def full_hook(module, module_input, module_output):
            return hook(module_output, name=self.name)
        if dir=='fwd':
            handle = self.register_forward_hook(full_hook)
            self.fwd_hooks.append(handle)
        elif dir=='bwd':
            handle = self.register_backward_hook(full_hook)
            self.bwd_hooks.append(handle)
        else:
            raise ValueError(f"Invalid direction {dir}")
    
    def remove_hooks(self, dir='fwd'):
        if (dir=='fwd') or (dir=='both'):
            for hook in self.fwd_hooks:
                hook.remove()
            self.fwd_hooks = []
        if (dir=='bwd') or (dir=='both'):
            for hook in self.bwd_hooks:
                hook.remove()
            self.bwd_hooks = []
        if dir not in ['fwd', 'bwd', 'both']:
            raise ValueError(f"Invalid direction {dir}")
    
    def forward(self, x):
        return x

# %% ../transformer.ipynb 5
class Embed(nn.Module):
    '''Define network architecture
    I defined my own transformer from scratch so I'd fully understand each component 
    - I expect this wasn't necessary or particularly important, and a bunch of this replicates existing PyTorch functionality
    '''
    def __init__(self, d_vocab, d_model):
        super().__init__()
        self.W_E = nn.Parameter(torch.randn(d_model, d_vocab)/np.sqrt(d_model))
    
    def forward(self, x):
        return torch.einsum('dbp -> bpd', self.W_E[:, x])

#| export
class Unembed(nn.Module):
    def __init__(self, d_vocab, d_model):
        super().__init__()
        self.W_U = nn.Parameter(torch.randn(d_model, d_vocab)/np.sqrt(d_vocab))
    
    def forward(self, x):
        return (x @ self.W_U)

#| export
class PosEmbed(nn.Module):
    def __init__(self, max_ctx, d_model):
        super().__init__()
        self.W_pos = nn.Parameter(torch.randn(max_ctx, d_model)/np.sqrt(d_model))
    
    def forward(self, x):
        return x+self.W_pos[:x.shape[-2]]

#| export
class LayerNorm(nn.Module):
    def __init__(self, d_model, epsilon = 1e-4, model=[None]):
        super().__init__()
        self.model = model
        self.w_ln = nn.Parameter(torch.ones(d_model))
        self.b_ln = nn.Parameter(torch.zeros(d_model))
        self.epsilon = epsilon
    
    def forward(self, x):
        if self.model[0].use_ln:
            x = x - x.mean(axis=-1)[..., None]
            x = x / (x.std(axis=-1)[..., None] + self.epsilon)
            x = x * self.w_ln
            x = x + self.b_ln
            return x
        else:
            return x

#| export
class Attention(nn.Module):
    def __init__(self, d_model, num_heads, d_head, n_ctx, model):
        super().__init__()
        self.model = model
        self.W_K = nn.Parameter(torch.randn(num_heads, d_head, d_model)/np.sqrt(d_model))
        self.W_Q = nn.Parameter(torch.randn(num_heads, d_head, d_model)/np.sqrt(d_model))
        self.W_V = nn.Parameter(torch.randn(num_heads, d_head, d_model)/np.sqrt(d_model))
        self.W_O = nn.Parameter(torch.randn(d_model, d_head * num_heads)/np.sqrt(d_model))
        self.register_buffer('mask', torch.tril(torch.ones((n_ctx, n_ctx))))
        self.d_head = d_head
        self.hook_k = HookPoint()
        self.hook_q = HookPoint()
        self.hook_v = HookPoint()
        self.hook_z = HookPoint()
        self.hook_attn = HookPoint()
        self.hook_attn_pre = HookPoint()

    def forward(self, x):
        k = self.hook_k(torch.einsum('ihd,bpd->biph', self.W_K, x))
        q = self.hook_q(torch.einsum('ihd,bpd->biph', self.W_Q, x))
        v = self.hook_v(torch.einsum('ihd,bpd->biph', self.W_V, x))
        attn_scores_pre = torch.einsum('biph,biqh->biqp', k, q)
        attn_scores_masked = torch.tril(attn_scores_pre) - 1e10 * (1 - self.mask[:x.shape[-2], :x.shape[-2]])
        attn_matrix = self.hook_attn(F.softmax(self.hook_attn_pre(attn_scores_masked/np.sqrt(self.d_head)), dim=-1))
        z = self.hook_z(torch.einsum('biph,biqp->biqh', v, attn_matrix))
        z_flat = einops.rearrange(z, 'b i q h -> b q (i h)')
        out = torch.einsum('df,bqf->bqd', self.W_O, z_flat)
        return out

#| export
class MLP(nn.Module):
    def __init__(self, d_model, d_mlp, act_type, model):
        super().__init__()
        self.model = model
        self.W_in = nn.Parameter(torch.randn(d_mlp, d_model)/np.sqrt(d_model))
        self.b_in = nn.Parameter(torch.zeros(d_mlp))
        self.W_out = nn.Parameter(torch.randn(d_model, d_mlp)/np.sqrt(d_model))
        self.b_out = nn.Parameter(torch.zeros(d_model))
        self.act_type = act_type
        # self.ln = LayerNorm(d_mlp, model=self.model)
        self.hook_pre = HookPoint()
        self.hook_post = HookPoint()
        assert act_type in ['ReLU', 'GeLU']
        
    def forward(self, x):
        x = self.hook_pre(torch.einsum('md,bpd->bpm', self.W_in, x) + self.b_in)
        if self.act_type=='ReLU':
            x = F.relu(x)
        elif self.act_type=='GeLU':
            x = F.gelu(x)
        x = self.hook_post(x)
        x = torch.einsum('dm,bpm->bpd', self.W_out, x) + self.b_out
        return x

# export
class TransformerBlock(nn.Module):
    def __init__(self, d_model, d_mlp, d_head, num_heads, n_ctx, act_type, model):
        super().__init__()
        self.model = model
        # self.ln1 = LayerNorm(d_model, model=self.model)
        self.attn = Attention(d_model, num_heads, d_head, n_ctx, model=self.model)
        # self.ln2 = LayerNorm(d_model, model=self.model)
        self.mlp = MLP(d_model, d_mlp, act_type, model=self.model)
        self.hook_attn_out = HookPoint()
        self.hook_mlp_out = HookPoint()
        self.hook_resid_pre = HookPoint()
        self.hook_resid_mid = HookPoint()
        self.hook_resid_post = HookPoint()
    
    def forward(self, x):
        x = self.hook_resid_mid(x + self.hook_attn_out(self.attn((self.hook_resid_pre(x)))))
        x = self.hook_resid_post(x + self.hook_mlp_out(self.mlp((x))))
        return x

#| export
class Transformer(nn.Module):
    def __init__(self, num_layers, d_vocab, d_model, d_mlp, d_head, num_heads, n_ctx, act_type, use_cache=False, use_ln=True):
        super().__init__()
        self.cache = {}
        self.use_cache = use_cache

        self.embed = Embed(d_vocab, d_model)
        self.pos_embed = PosEmbed(n_ctx, d_model)
        self.blocks = nn.ModuleList([TransformerBlock(d_model, d_mlp, d_head, num_heads, n_ctx, act_type, model=[self]) for i in range(num_layers)])
        # self.ln = LayerNorm(d_model, model=[self])
        self.unembed = Unembed(d_vocab, d_model)
        self.use_ln = use_ln

        for name, module in self.named_modules():
            if type(module)==HookPoint:
                module.give_name(name)
    
    def forward(self, x):
        x = self.embed(x)
        x = self.pos_embed(x)
        for block in self.blocks:
            x = block(x)
        # x = self.ln(x)
        x = self.unembed(x)
        return x

    def set_use_cache(self, use_cache):
        self.use_cache = use_cache
    
    def hook_points(self):
        return [module for name, module in self.named_modules() if 'hook' in name]

    def remove_all_hooks(self):
        for hp in self.hook_points():
            hp.remove_hooks('fwd')
            hp.remove_hooks('bwd')
    
    def cache_all(self, cache, incl_bwd=False):
        # Caches all activations wrapped in a HookPoint
        def save_hook(tensor, name):
            cache[name] = tensor.detach()
        def save_hook_back(tensor, name):
            cache[name+'_grad'] = tensor[0].detach()
        for hp in self.hook_points():
            hp.add_hook(save_hook, 'fwd')
            if incl_bwd:
                hp.add_hook(save_hook_back, 'bwd')

# %% ../transformer.ipynb 6
import numpy as np
import random
import torch

lr=1e-3 #@param
weight_decay = 1.0 #@param
p=113 #@param
d_model = 128 #@param
fn_name = 'add' #@param ['add', 'subtract', 'x2xyy2','rand']
frac_train = 0.3 #@param
num_epochs = 50000 #@param
save_models = False #@param
save_every = 100 #@param
# Stop training when test loss is <stopping_thresh
stopping_thresh = -1 #@param
seed = 0 #@param

num_layers = 1
batch_style = 'full'
d_vocab = p+1
n_ctx = 3
d_mlp = 4*d_model
num_heads = 4
assert d_model % num_heads == 0
d_head = d_model//num_heads
act_type = 'ReLU' #@param ['ReLU', 'GeLU']
# batch_size = 512
use_ln = False
random_answers = np.random.randint(low=0, high=p, size=(p, p))
fns_dict = {'add': lambda x,y:(x+y)%p, 'subtract': lambda x,y:(x-y)%p, 'x2xyy2':lambda x,y:(x**2+x*y+y**2)%p, 'rand':lambda x,y:random_answers[x][y]}
fn = fns_dict[fn_name]

def gen_train_test(frac_train, num, seed=0):
    # Generate train and test split
    pairs = [(i, j, num) for i in range(num) for j in range(num)]
    random.seed(seed)
    random.shuffle(pairs)
    div = int(frac_train*len(pairs))
    return pairs[:div], pairs[div:]

train, test = gen_train_test(frac_train, p, seed)
print(f'{len(train)=}, {len(test)=}')

def train_model():
    # Creates an array of Boolean indices according to whether each data point is in 
    # train or test
    # Used to index into the big batch of all possible data
    is_train = []
    is_test = []
    for x in range(p):
        for y in range(p):
            if (x, y, 113) in train:
                is_train.append(True)
                is_test.append(False)
            else:
                is_train.append(False)
                is_test.append(True)
    is_train = np.array(is_train)
    is_test = np.array(is_test)

    model = Transformer(num_layers=num_layers, d_vocab=d_vocab, d_model=d_model, d_mlp=d_mlp, d_head=d_head, num_heads=num_heads, n_ctx=n_ctx, act_type=act_type, use_cache=False, use_ln=use_ln)
    model.to('cuda')
    optimizer = optim.AdamW(model.parameters(), lr=lr, weight_decay=weight_decay, betas=(0.9, 0.98))
    scheduler = optim.lr_scheduler.LambdaLR(optimizer, lambda step: min(step/10, 1))
    run_name = f"grok_{int(time.time())}"
    print(f'Run name {run_name}')
    if save_models:
        os.mkdir(root/run_name)
        save_dict = {'model':model.state_dict(), 'train_data':train, 'test_data':test}
        torch.save(save_dict, root/run_name/'init.pth')
    train_losses = []
    test_losses = []
    for epoch in range(num_epochs):
        train_loss = full_loss(model, train)
        test_loss = full_loss(model, test)
        train_losses.append(train_loss.item())
        test_losses.append(test_loss.item())
        if epoch%100 == 0: print(f"{epoch}_{np.log(train_loss.item()):.4f}_{np.log(test_loss.item()):.4f}")#_{train_acc.item():.4f}_{test_acc.item():.4f}")
        train_loss.backward()
        optimizer.step()
        scheduler.step()
        optimizer.zero_grad()
        if test_loss.item() < stopping_thresh:
            break
        if (save_models) and (epoch%save_every == 0):
            if test_loss.item() < stopping_thresh:
                break
            save_dict = {
                'model': model.state_dict(),
                'optimizer': optimizer.state_dict(),
                'scheduler': scheduler.state_dict(),
                'train_loss': train_loss,
                'test_loss': test_loss,
                'epoch': epoch,
            }
            torch.save(save_dict, root/run_name/f"{epoch}.pth")
            print(f"Saved model to {root/run_name/f'{epoch}.pth'}")
    if not save_models:
        os.mkdir(root/run_name)
    save_dict = {
        'model': model.state_dict(),
        'optimizer': optimizer.state_dict(),
        'scheduler': scheduler.state_dict(),
        'train_loss': train_loss,
        'test_loss': test_loss,
        'train_losses': train_losses,
        'test_losses': test_losses,
        'epoch': epoch,
    }
    torch.save(save_dict, root/run_name/f"final.pth")
    print(f"Saved model to {root/run_name/f'final.pth'}")
    lines([train_losses, test_losses], labels=['train', 'test'], log_y=True)

    # save_models = False



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/grokking',
                'doc_host': 'https://smithjessk.github.io',
                'git_url': 'https://github.com/smithjessk/grokking',
                'lib_path': 'grokking'},
  'syms': { 'grokking.helpers': { 'grokking.helpers.animate_lines': ('helpers.html#animate_lines', 'grokking/helpers.py'),
                                  'grokking.helpers.animate_multi_lines': ('helpers.html#animate_multi_lines', 'grokking/helpers.py'),
                                  'grokking.helpers.animate_scatter': ('helpers.html#animate_scatter', 'grokking/helpers.py'),
                                  'grokking.helpers.cos': ('helpers.html#cos', 'grokking/helpers.py'),
                                  'grokking.helpers.cpu_aware_load_at_root': ('helpers.html#cpu_aware_load_at_root', 'grokking/helpers.py'),
                                  'grokking.helpers.cross_entropy_high_precision': ( 'helpers.html#cross_entropy_high_precision',
                                                                                     'grokking/helpers.py'),
                                  'grokking.helpers.cuda_memory': ('helpers.html#cuda_memory', 'grokking/helpers.py'),
                                  'grokking.helpers.download_model_from_drive': ( 'helpers.html#download_model_from_drive',
                                                                                  'grokking/helpers.py'),
                                  'grokking.helpers.extract_freq_2d': ('helpers.html#extract_freq_2d', 'grokking/helpers.py'),
                                  'grokking.helpers.full_loss': ('helpers.html#full_loss', 'grokking/helpers.py'),
                                  'grokking.helpers.get_cov': ('helpers.html#get_cov', 'grokking/helpers.py'),
                                  'grokking.helpers.imshow': ('helpers.html#imshow', 'grokking/helpers.py'),
                                  'grokking.helpers.imshow_fourier': ('helpers.html#imshow_fourier', 'grokking/helpers.py'),
                                  'grokking.helpers.install_dependencies': ('helpers.html#install_dependencies', 'grokking/helpers.py'),
                                  'grokking.helpers.is_close': ('helpers.html#is_close', 'grokking/helpers.py'),
                                  'grokking.helpers.is_inside_colab': ('helpers.html#is_inside_colab', 'grokking/helpers.py'),
                                  'grokking.helpers.line': ('helpers.html#line', 'grokking/helpers.py'),
                                  'grokking.helpers.line_marker': ('helpers.html#line_marker', 'grokking/helpers.py'),
                                  'grokking.helpers.lines': ('helpers.html#lines', 'grokking/helpers.py'),
                                  'grokking.helpers.load_5_digit_addition_finite': ( 'helpers.html#load_5_digit_addition_finite',
                                                                                     'grokking/helpers.py'),
                                  'grokking.helpers.load_5_digit_addition_infinite': ( 'helpers.html#load_5_digit_addition_infinite',
                                                                                       'grokking/helpers.py'),
                                  'grokking.helpers.load_finite_data_losses': ( 'helpers.html#load_finite_data_losses',
                                                                                'grokking/helpers.py'),
                                  'grokking.helpers.load_induction_head_finite': ( 'helpers.html#load_induction_head_finite',
                                                                                   'grokking/helpers.py'),
                                  'grokking.helpers.load_induction_head_infinite': ( 'helpers.html#load_induction_head_infinite',
                                                                                     'grokking/helpers.py'),
                                  'grokking.helpers.load_infinite_data_losses': ( 'helpers.html#load_infinite_data_losses',
                                                                                  'grokking/helpers.py'),
                                  'grokking.helpers.load_mod_addition_frac_train_sweep': ( 'helpers.html#load_mod_addition_frac_train_sweep',
                                                                                           'grokking/helpers.py'),
                                  'grokking.helpers.load_no_wd_width_scan': ('helpers.html#load_no_wd_width_scan', 'grokking/helpers.py'),
                                  'grokking.helpers.mod_div': ('helpers.html#mod_div', 'grokking/helpers.py'),
                                  'grokking.helpers.normalize': ('helpers.html#normalize', 'grokking/helpers.py'),
                                  'grokking.helpers.run_shell_command_as_python': ( 'helpers.html#run_shell_command_as_python',
                                                                                    'grokking/helpers.py'),
                                  'grokking.helpers.scatter': ('helpers.html#scatter', 'grokking/helpers.py'),
                                  'grokking.helpers.test_logits': ('helpers.html#test_logits', 'grokking/helpers.py'),
                                  'grokking.helpers.to_numpy': ('helpers.html#to_numpy', 'grokking/helpers.py'),
                                  'grokking.helpers.unflatten_first': ('helpers.html#unflatten_first', 'grokking/helpers.py')},
            'grokking.setup': {},
            'grokking.transformers': { 'grokking.transformers.Attention': ('transformer.html#attention', 'grokking/transformers.py'),
                                       'grokking.transformers.Attention.__init__': ( 'transformer.html#__init__',
                                                                                     'grokking/transformers.py'),
                                       'grokking.transformers.Attention.forward': ('transformer.html#forward', 'grokking/transformers.py'),
                                       'grokking.transformers.Embed': ('transformer.html#embed', 'grokking/transformers.py'),
                                       'grokking.transformers.Embed.__init__': ('transformer.html#__init__', 'grokking/transformers.py'),
                                       'grokking.transformers.Embed.forward': ('transformer.html#forward', 'grokking/transformers.py'),
                                       'grokking.transformers.HookPoint': ('transformer.html#hookpoint', 'grokking/transformers.py'),
                                       'grokking.transformers.HookPoint.__init__': ( 'transformer.html#__init__',
                                                                                     'grokking/transformers.py'),
                                       'grokking.transformers.HookPoint.add_hook': ( 'transformer.html#add_hook',
                                                                                     'grokking/transformers.py'),
                                       'grokking.transformers.HookPoint.forward': ('transformer.html#forward', 'grokking/transformers.py'),
                                       'grokking.transformers.HookPoint.give_name': ( 'transformer.html#give_name',
                                                                                      'grokking/transformers.py'),
                                       'grokking.transformers.HookPoint.remove_hooks': ( 'transformer.html#remove_hooks',
                                                                                         'grokking/transformers.py'),
                                       'grokking.transformers.LayerNorm': ('transformer.html#layernorm', 'grokking/transformers.py'),
                                       'grokking.transformers.LayerNorm.__init__': ( 'transformer.html#__init__',
                                                                                     'grokking/transformers.py'),
                                       'grokking.transformers.LayerNorm.forward': ('transformer.html#forward', 'grokking/transformers.py'),
                                       'grokking.transformers.MLP': ('transformer.html#mlp', 'grokking/transformers.py'),
                                       'grokking.transformers.MLP.__init__': ('transformer.html#__init__', 'grokking/transformers.py'),
                                       'grokking.transformers.MLP.forward': ('transformer.html#forward', 'grokking/transformers.py'),
                                       'grokking.transformers.PosEmbed': ('transformer.html#posembed', 'grokking/transformers.py'),
                                       'grokking.transformers.PosEmbed.__init__': ('transformer.html#__init__', 'grokking/transformers.py'),
                                       'grokking.transformers.PosEmbed.forward': ('transformer.html#forward', 'grokking/transformers.py'),
                                       'grokking.transformers.Transformer': ('transformer.html#transformer', 'grokking/transformers.py'),
                                       'grokking.transformers.Transformer.__init__': ( 'transformer.html#__init__',
                                                                                       'grokking/transformers.py'),
                                       'grokking.transformers.Transformer.cache_all': ( 'transformer.html#cache_all',
                                                                                        'grokking/transformers.py'),
                                       'grokking.transformers.Transformer.forward': ( 'transformer.html#forward',
                                                                                      'grokking/transformers.py'),
                                       'grokking.transformers.Transformer.hook_points': ( 'transformer.html#hook_points',
                                                                                          'grokking/transformers.py'),
                                       'grokking.transformers.Transformer.remove_all_hooks': ( 'transformer.html#remove_all_hooks',
                                                                                               'grokking/transformers.py'),
                                       'grokking.transformers.Transformer.set_use_cache': ( 'transformer.html#set_use_cache',
                                                                                            'grokking/transformers.py'),
                                       'grokking.transformers.TransformerBlock': ( 'transformer.html#transformerblock',
                                                                                   'grokking/transformers.py'),
                                       'grokking.transformers.TransformerBlock.__init__': ( 'transformer.html#__init__',
                                                                                            'grokking/transformers.py'),
                                       'grokking.transformers.TransformerBlock.forward': ( 'transformer.html#forward',
                                                                                           'grokking/transformers.py'),
                                       'grokking.transformers.Unembed': ('transformer.html#unembed', 'grokking/transformers.py'),
                                       'grokking.transformers.Unembed.__init__': ('transformer.html#__init__', 'grokking/transformers.py'),
                                       'grokking.transformers.Unembed.forward': ('transformer.html#forward', 'grokking/transformers.py'),
                                       'grokking.transformers.gen_train_test': ( 'transformer.html#gen_train_test',
                                                                                 'grokking/transformers.py'),
                                       'grokking.transformers.train_model': ('transformer.html#train_model', 'grokking/transformers.py')}}}


# AUTOGENERATED! DO NOT EDIT! File to edit: ../helpers.ipynb.

# %% auto 0
__all__ = ['use_drive', 'root', 'imshow', 'imshow_div', 'inputs_heatmap', 'run_shell_command_as_python', 'is_inside_colab',
           'install_dependencies', 'download_model_from_drive', 'cuda_memory', 'cross_entropy_high_precision',
           'full_loss', 'test_logits', 'to_numpy', 'scatter', 'line', 'lines', 'line_marker', 'animate_lines',
           'imshow_fourier', 'animate_multi_lines', 'animate_scatter', 'unflatten_first', 'cos', 'mod_div', 'normalize',
           'extract_freq_2d', 'get_cov', 'is_close', 'cpu_aware_load_at_root', 'load_mod_addition_frac_train_sweep',
           'load_5_digit_addition_infinite', 'load_5_digit_addition_finite', 'load_induction_head_finite',
           'load_induction_head_infinite', 'load_infinite_data_losses', 'load_finite_data_losses',
           'load_no_wd_width_scan']

# %% ../helpers.ipynb 2
def run_shell_command_as_python(shell):
    '''helpful for python functions; thanks https://stackoverflow.com/questions/70068720/jupyter-shell-commands-in-a-function'''
    from IPython import get_ipython
    ipython = get_ipython()
    code = ipython.transform_cell(f'!{shell}')
    print(f'Executing {code}')
    exec(code)

# %% ../helpers.ipynb 3
def is_inside_colab():
    try:
        import google.colab
        return True
    except:
        return False

# %% ../helpers.ipynb 5
def install_dependencies():
    # TODO how to make this run at the right times?
    run_shell_command_as_python("nvidia-smi") # TODO what if this isn't available? maybe this is just for the main notebook?
    run_shell_command_as_python("pip install einops matplotlib pandas plotly")

use_drive = False #@param

# Import stuff
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import numpy as np
import einops
import tqdm.notebook as tqdm

import random
import time

# TODO from google.colab import drive
from pathlib import Path
import pickle
import os

import matplotlib.pyplot as plt

import plotly.express as px
import plotly.io as pio
if is_inside_colab():
    pio.renderers.default = "colab"
else:
    pio.renderers.default = "vscode"
import plotly.graph_objects as go

from torch.utils.data import DataLoader

from functools import *
import pandas as pd
import gc

import itertools

# %% ../helpers.ipynb 7
root = Path('./saved_runs')

#| export
def download_model_from_drive():
    # TODO can we use google drive for these too?
    try:
        run_shell_command_as_python(f"git clone https://github.com/neelnanda-io/Grokking.git {root}")
        run_shell_command_as_python(f"mv {root}/saved_runs/* {root} || true") # in case they're already downloaded, TODO clean up   
    except:
        pass
    large_root = Path('./large_files')
    os.makedirs(large_root, exist_ok=True)

    output_file = large_root/'full_run_data.pth'
    if not output_file.exists():
        run_shell_command_as_python("pip install gdown")
        run_shell_command_as_python(f"gdown 12pmgxpTHLDzSNMbMCuAMXP1lE_XiCQRy -O {large_root}/full_run_data.pth")
    else:
        print('saved runs already exist, skipping...')

    if use_drive:
        print('setting up drive...')
        drive.mount('./gdrive')
        drive_root = Path('./gdrive/MyDrive/Colab Notebooks/Grok/')

# %% ../helpers.ipynb 9
def cuda_memory():
    print(torch.cuda.memory_allocated()/1e9)

#| export
def cross_entropy_high_precision(logits, labels):
    # Shapes: batch x vocab, batch
    # Cast logits to float64 because log_softmax has a float32 underflow on overly 
    # confident data and can only return multiples of 1.2e-7 (the smallest float x
    # such that 1+x is different from 1 in float32). This leads to loss spikes 
    # and dodgy gradients
    logprobs = F.log_softmax(logits.to(torch.float64), dim=-1)
    prediction_logprobs = torch.gather(logprobs, index=labels[:, None], dim=-1)
    loss = -torch.mean(prediction_logprobs)
    return loss

#| export
def full_loss(model, data):
    # Take the final position only
    logits = model(data)[:, -1]
    labels = torch.tensor([fn(i, j) for i, j, _ in data]).to('cuda')
    return cross_entropy_high_precision(logits, labels)

#| export
def test_logits(logits, bias_correction=False, original_logits=None, mode='all'):
    # Calculates cross entropy loss of logits representing a batch of all p^2 
    # possible inputs
    # Batch dimension is assumed to be first
    if logits.shape[1]==p*p:
        logits = logits.T
    if logits.shape==torch.Size([p*p, p+1]):
        logits = logits[:, :-1]
    logits = logits.reshape(p*p, p)
    if bias_correction:
        # Applies bias correction - we correct for any missing bias terms, 
        # independent of the input, by centering the new logits along the batch 
        # dimension, and then adding the average original logits across all inputs
        logits = einops.reduce(original_logits - logits, 'batch ... -> ...', 'mean') + logits
    if mode=='train':
        return cross_entropy_high_precision(logits[is_train], labels[is_train])
    elif mode=='test':
        return cross_entropy_high_precision(logits[is_test], labels[is_test])
    elif mode=='all':
        return cross_entropy_high_precision(logits, labels)

# %% ../helpers.ipynb 12
def to_numpy(tensor, flat=False):
    if type(tensor)!=torch.Tensor:
        return tensor
    if flat:
        return tensor.flatten().detach().cpu().numpy()
    else:
        return tensor.detach().cpu().numpy()

def imshow(tensor, xaxis=None, yaxis=None, animation_name='Snapshot', **kwargs):
    if tensor.shape[0]==p*p:
        tensor = unflatten_first(tensor)
    tensor = torch.squeeze(tensor)
    px.imshow(to_numpy(tensor, flat=False), 
              labels={'x':xaxis, 'y':yaxis, 'animation_name':animation_name}, 
              **kwargs).show()

# Set default colour scheme
#| export
imshow = partial(imshow, color_continuous_scale='Blues')

# Creates good defaults for showing divergent colour scales (ie with both 
# positive and negative values, where 0 is white)
imshow_div = partial(imshow, color_continuous_scale='RdBu', color_continuous_midpoint=0.0)

# Presets a bunch of defaults to imshow to make it suitable for showing heatmaps 
# of activations with x axis being input 1 and y axis being input 2.
#| export
inputs_heatmap = partial(imshow, xaxis='Input 1', yaxis='Input 2', color_continuous_scale='RdBu', color_continuous_midpoint=0.0)


def scatter(x, y, **kwargs):
    px.scatter(x=to_numpy(x, flat=True), y=to_numpy(y, flat=True), **kwargs).show()

# %% ../helpers.ipynb 13
def line(x, y=None, hover=None, xaxis='', yaxis='', **kwargs):
    if type(y)==torch.Tensor:
        y = to_numpy(y, flat=True)
    if type(x)==torch.Tensor:
        x=to_numpy(x, flat=True)
    fig = px.line(x, y=y, hover_name=hover, **kwargs)
    fig.update_layout(xaxis_title=xaxis, yaxis_title=yaxis)
    fig.show()

# %% ../helpers.ipynb 14
def lines(lines_list, x=None, mode='lines', labels=None, xaxis='', yaxis='', title = '', log_y=False, hover=None, **kwargs):
    '''Helper function to plot multiple lines'''
    if type(lines_list)==torch.Tensor:
        lines_list = [lines_list[i] for i in range(lines_list.shape[0])]
    if x is None:
        x=np.arange(len(lines_list[0]))
    fig = go.Figure(layout={'title':title})
    fig.update_xaxes(title=xaxis)
    fig.update_yaxes(title=yaxis)
    for c, line in enumerate(lines_list):
        if type(line)==torch.Tensor:
            line = to_numpy(line)
        if labels is not None:
            label = labels[c]
        else:
            label = c
        fig.add_trace(go.Scatter(x=x, y=line, mode=mode, name=label, hovertext=hover, **kwargs))
    if log_y:
        fig.update_layout(yaxis_type="log")
    fig.show()

def line_marker(x, **kwargs):
    lines([x], mode='lines+markers', **kwargs)

def animate_lines(lines_list, snapshot_index = None, snapshot='snapshot', hover=None, xaxis='x', yaxis='y', **kwargs):
    if type(lines_list)==list:
        lines_list = torch.stack(lines_list, axis=0)
    lines_list = to_numpy(lines_list, flat=False)
    if snapshot_index is None:
        snapshot_index = np.arange(lines_list.shape[0])
    if hover is not None:
        hover = [i for j in range(len(snapshot_index)) for i in hover]
    print(lines_list.shape)
    rows=[]
    for i in range(lines_list.shape[0]):
        for j in range(lines_list.shape[1]):
            rows.append([lines_list[i][j], snapshot_index[i], j])
    df = pd.DataFrame(rows, columns=[yaxis, snapshot, xaxis])
    px.line(df, x=xaxis, y=yaxis, animation_frame=snapshot, range_y=[lines_list.min(), lines_list.max()], hover_name=hover,**kwargs).show()

def imshow_fourier(tensor, title='', animation_name='snapshot', facet_labels=[], **kwargs):
    # Set nice defaults for plotting functions in the 2D fourier basis
    # tensor is assumed to already be in the Fourier Basis
    if tensor.shape[0]==p*p:
        tensor = unflatten_first(tensor)
    tensor = torch.squeeze(tensor)
    fig=px.imshow(to_numpy(tensor),
            x=fourier_basis_names, 
            y=fourier_basis_names, 
            labels={'x':'x Component', 
                    'y':'y Component', 
                    'animation_frame':animation_name},
            title=title,
            color_continuous_midpoint=0., 
            color_continuous_scale='RdBu', 
            **kwargs)
    fig.update(data=[{'hovertemplate':"%{x}x * %{y}y<br>Value:%{z:.4f}"}])
    if facet_labels:
        for i, label in enumerate(facet_labels):
            fig.layout.annotations[i]['text'] = label
    fig.show()

def animate_multi_lines(lines_list, y_index=None, snapshot_index = None, snapshot='snapshot', hover=None, swap_y_animate=False, **kwargs):
    # Can plot an animation of lines with multiple lines on the plot.
    if type(lines_list)==list:
        lines_list = torch.stack(lines_list, axis=0)
    lines_list = to_numpy(lines_list, flat=False)
    if swap_y_animate:
        lines_list = lines_list.transpose(1, 0, 2)
    if snapshot_index is None:
        snapshot_index = np.arange(lines_list.shape[0])
    if y_index is None:
        y_index = [str(i) for i in range(lines_list.shape[1])]
    if hover is not None:
        hover = [i for j in range(len(snapshot_index)) for i in hover]
    print(lines_list.shape)
    rows=[]
    for i in range(lines_list.shape[0]):
        for j in range(lines_list.shape[2]):
            rows.append(list(lines_list[i, :, j])+[snapshot_index[i], j])
    df = pd.DataFrame(rows, columns=y_index+[snapshot, 'x'])
    px.line(df, x='x', y=y_index, animation_frame=snapshot, range_y=[lines_list.min(), lines_list.max()], hover_name=hover, **kwargs).show()

def animate_scatter(lines_list, snapshot_index = None, snapshot='snapshot', hover=None, yaxis='y', xaxis='x', color=None, color_name = 'color', **kwargs):
    # Can plot an animated scatter plot
    # lines_list has shape snapshot x 2 x line
    if type(lines_list)==list:
        lines_list = torch.stack(lines_list, axis=0)
    lines_list = to_numpy(lines_list, flat=False)
    if snapshot_index is None:
        snapshot_index = np.arange(lines_list.shape[0])
    if hover is not None:
        hover = [i for j in range(len(snapshot_index)) for i in hover]
    if color is None:
        color = np.ones(lines_list.shape[-1])
    if type(color)==torch.Tensor:
        color = to_numpy(color)
    if len(color.shape)==1:
        color = einops.repeat(color, 'x -> snapshot x', snapshot=lines_list.shape[0])
    print(lines_list.shape)
    rows=[]
    for i in range(lines_list.shape[0]):
        for j in range(lines_list.shape[2]):
            rows.append([lines_list[i, 0, j].item(), lines_list[i, 1, j].item(), snapshot_index[i], color[i, j]])
    print([lines_list[:, 0].min(), lines_list[:, 0].max()])
    print([lines_list[:, 1].min(), lines_list[:, 1].max()])
    df = pd.DataFrame(rows, columns=[xaxis, yaxis, snapshot, color_name])
    px.scatter(df, x=xaxis, y=yaxis, animation_frame=snapshot, range_x=[lines_list[:, 0].min(), lines_list[:, 0].max()], range_y=[lines_list[:, 1].min(), lines_list[:, 1].max()], hover_name=hover, color=color_name, **kwargs).show()

# %% ../helpers.ipynb 15
def unflatten_first(tensor):
    if tensor.shape[0]==p*p:
        return einops.rearrange(tensor, '(x y) ... -> x y ...', x=p, y=p)
    else: 
        return tensor
def cos(x, y):
    return (x.dot(y))/x.norm()/y.norm()
def mod_div(a, b):
    return (a*pow(b, p-2, p))%p
def normalize(tensor, axis=0):
    return tensor/(tensor).pow(2).sum(keepdim=True, axis=axis).sqrt()
def extract_freq_2d(tensor, freq):
    # Takes in a pxpx... or batch x ... tensor, returns a 3x3x... tensor of the 
    # Linear and quadratic terms of frequency freq
    tensor = unflatten_first(tensor)
    # Extracts the linear and quadratic terms corresponding to frequency freq
    index_1d = [0, 2*freq-1, 2*freq]
    # Some dumb manipulation to use fancy array indexing rules
    # Gets the rows and columns in index_1d
    return tensor[[[i]*3 for i in index_1d], [index_1d]*3]
def get_cov(tensor, norm=True):
    # Calculate covariance matrix
    if norm:
        tensor = normalize(tensor, axis=1)
    return tensor @ tensor.T
def is_close(a, b):
    return ((a-b).pow(2).sum()/(a.pow(2).sum().sqrt())/(b.pow(2).sum().sqrt())).item()

# %% ../helpers.ipynb 17
def cpu_aware_load_at_root(path):
    path = root / path
    if torch.cuda.is_available():
        return torch.load(path)
    else:
        return torch.load(path, map_location=torch.device('cpu'))

#| export
def load_mod_addition_frac_train_sweep():
    return cpu_aware_load_at_root('mod_addition_frac_train_sweep.pth')

#| export
def load_5_digit_addition_infinite():
    return cpu_aware_load_at_root('5_digit_addition_infinite.pth')

#| export
def load_5_digit_addition_finite():
    return cpu_aware_load_at_root('5_digit_addition_finite.pth')

#| export
def load_induction_head_finite():
    return cpu_aware_load_at_root('induction_head_finite.pth')

#| export
def load_induction_head_infinite():
    return cpu_aware_load_at_root('induction_head_infinite.pth')

#| export
def load_infinite_data_losses():
    return cpu_aware_load_at_root('skip_trigram_infinite.pth')

#| export
def load_finite_data_losses():
    return cpu_aware_load_at_root('skip_trigram_finite.pth')

#| export
def load_no_wd_width_scan():
    return cpu_aware_load_at_root('no_wd_width_scan.pth')

